@@ -1,8 +1,6 @@
-// Emacs style mode select   -*- C++ -*- 
-//-----------------------------------------------------------------------------
 //
 // Copyright(C) 1993-1996 Id Software, Inc.
-// Copyright(C) 2005 Simon Howard
+// Copyright(C) 2005-2014 Simon Howard
 //
 // This program is free software; you can redistribute it and/or
 // modify it under the terms of the GNU General Public License
@@ -14,29 +12,22 @@
 // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 // GNU General Public License for more details.
 //
-// You should have received a copy of the GNU General Public License
-// along with this program; if not, write to the Free Software
-// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-// 02111-1307, USA.
-//
 // DESCRIPTION:
 //	System interface for music.
 //
-//-----------------------------------------------------------------------------
 
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
-#include "doomdef.h"
 #include "memio.h"
 #include "mus2mid.h"
 
 #include "deh_main.h"
+#include "i_sound.h"
 #include "i_swap.h"
 #include "m_misc.h"
-#include "s_sound.h"
 #include "w_wad.h"
 #include "z_zone.h"
 
@@ -112,11 +103,6 @@ typedef struct
     // Track iterator used to read new events.
 
     midi_track_iter_t *iter;
-
-    // Tempo control variables
-
-    unsigned int ticks_per_beat;
-    unsigned int ms_per_beat;
 } opl_track_data_t;
 
 typedef struct opl_voice_s opl_voice_t;
@@ -333,6 +319,11 @@ static unsigned int num_tracks = 0;
 static unsigned int running_tracks = 0;
 static boolean song_looping;
 
+// Tempo control variables
+
+static unsigned int ticks_per_beat;
+static unsigned int us_per_beat;
+
 // Mini-log of recently played percussion instruments:
 
 static uint8_t last_perc[PERCUSSION_LOG_LEN];
@@ -736,7 +727,7 @@ static unsigned int FrequencyForVoice(opl_voice_t *voice)
 
     gm_voice = &voice->current_instr->voices[voice->current_instr_voice];
 
-    if ((voice->current_instr->flags & GENMIDI_FLAG_FIXED) == 0)
+    if ((SHORT(voice->current_instr->flags) & GENMIDI_FLAG_FIXED) == 0)
     {
         note += (signed short) SHORT(gm_voice->base_note_offset);
     }
@@ -853,7 +844,7 @@ static void VoiceKeyOn(opl_channel_data_t *channel,
     // Work out the note to use.  This is normally the same as
     // the key, unless it is a fixed pitch instrument.
 
-    if ((instrument->flags & GENMIDI_FLAG_FIXED) != 0)
+    if ((SHORT(instrument->flags) & GENMIDI_FLAG_FIXED) != 0)
     {
         voice->note = instrument->fixed_note;
     }
@@ -890,12 +881,21 @@ static void KeyOnEvent(opl_track_data_t *track, midi_event_t *event)
            event->data.channel.param2);
 */
 
-    // The channel.
-
-    channel = &track->channels[event->data.channel.channel];
     key = event->data.channel.param1;
     volume = event->data.channel.param2;
 
+    // A volume of zero means key off. Some MIDI tracks, eg. the ones
+    // in AV.wad, use a second key on with a volume of zero to mean
+    // key off.
+    if (volume <= 0)
+    {
+        KeyOffEvent(track, event);
+        return;
+    }
+
+    // The channel.
+    channel = &track->channels[event->data.channel.channel];
+
     // Percussion channel (10) is treated differently.
 
     if (event->data.channel.channel == 9)
@@ -920,7 +920,7 @@ static void KeyOnEvent(opl_track_data_t *track, midi_event_t *event)
 
     VoiceKeyOn(channel, instrument, 0, key, volume);
 
-    if ((instrument->flags & GENMIDI_FLAG_2VOICE) != 0)
+    if ((SHORT(instrument->flags) & GENMIDI_FLAG_2VOICE) != 0)
     {
         VoiceKeyOn(channel, instrument, 1, key, volume);
     }
@@ -958,6 +958,21 @@ static void SetChannelVolume(opl_channel_data_t *channel, unsigned int volume)
     }
 }
 
+// Handler for the MIDI_CONTROLLER_ALL_NOTES_OFF channel event.
+static void AllNotesOff(opl_channel_data_t *channel, unsigned int param)
+{
+    unsigned int i;
+
+    for (i = 0; i < OPL_NUM_VOICES; ++i)
+    {
+        if (voices[i].channel == channel)
+        {
+            VoiceKeyOff(&voices[i]);
+            ReleaseVoice(&voices[i]);
+        }
+    }
+}
+
 static void ControllerEvent(opl_track_data_t *track, midi_event_t *event)
 {
     unsigned int controller;
@@ -981,6 +996,10 @@ static void ControllerEvent(opl_track_data_t *track, midi_event_t *event)
             SetChannelVolume(channel, param);
             break;
 
+        case MIDI_CONTROLLER_ALL_NOTES_OFF:
+            AllNotesOff(channel, param);
+            break;
+
         default:
 #ifdef OPL_MIDI_DEBUG
             fprintf(stderr, "Unknown MIDI controller type: %i\n", controller);
@@ -1013,10 +1032,19 @@ static void PitchBendEvent(opl_track_data_t *track, midi_event_t *event)
     }
 }
 
+static void MetaSetTempo(unsigned int tempo)
+{
+    OPL_AdjustCallbacks((float) us_per_beat / tempo);
+    us_per_beat = tempo;
+}
+
 // Process a meta event.
 
 static void MetaEvent(opl_track_data_t *track, midi_event_t *event)
 {
+    byte *data = event->data.meta.data;
+    unsigned int data_len = event->data.meta.length;
+
     switch (event->data.meta.type)
     {
         // Things we can just ignore.
@@ -1032,6 +1060,13 @@ static void MetaEvent(opl_track_data_t *track, midi_event_t *event)
         case MIDI_META_SEQUENCER_SPECIFIC:
             break;
 
+        case MIDI_META_SET_TEMPO:
+            if (data_len == 3)
+            {
+                MetaSetTempo((data[0] << 16) | (data[1] << 8) | data[2]);
+            }
+            break;
+
         // End of track - actually handled when we run out of events
         // in the track, see below.
 
@@ -1141,7 +1176,7 @@ static void TrackTimerCallback(void *arg)
 
         if (running_tracks <= 0 && song_looping)
         {
-            OPL_SetCallback(5, RestartSong, NULL);
+            OPL_SetCallback(5000, RestartSong, NULL);
         }
 
         return;
@@ -1155,19 +1190,17 @@ static void TrackTimerCallback(void *arg)
 static void ScheduleTrack(opl_track_data_t *track)
 {
     unsigned int nticks;
-    unsigned int ms;
-    static int total = 0;
+    uint64_t us;
 
-    // Get the number of milliseconds until the next event.
+    // Get the number of microseconds until the next event.
 
     nticks = MIDI_GetDeltaTime(track->iter);
-    ms = (nticks * track->ms_per_beat) / track->ticks_per_beat;
-    total += ms;
+    us = ((uint64_t) nticks * us_per_beat) / ticks_per_beat;
 
     // Set a timer to be invoked when the next event is
     // ready to play.
 
-    OPL_SetCallback(ms, TrackTimerCallback, track);
+    OPL_SetCallback(us, TrackTimerCallback, track);
 }
 
 // Initialize a channel.
@@ -1190,12 +1223,6 @@ static void StartTrack(midi_file_t *file, unsigned int track_num)
 
     track = &tracks[track_num];
     track->iter = MIDI_IterateTrack(file, track_num);
-    track->ticks_per_beat = MIDI_GetFileTimeDivision(file);
-
-    // Default is 120 bpm.
-    // TODO: this is wrong
-
-    track->ms_per_beat = 500 * 260;
 
     for (i=0; i<MIDI_CHANNELS_PER_TRACK; ++i)
     {
@@ -1209,7 +1236,7 @@ static void StartTrack(midi_file_t *file, unsigned int track_num)
 
 // Start playing a mid
 
-static void I_OPL_PlaySong(void *handle, int looping)
+static void I_OPL_PlaySong(void *handle, boolean looping)
 {
     midi_file_t *file;
     unsigned int i;
@@ -1229,6 +1256,13 @@ static void I_OPL_PlaySong(void *handle, int looping)
     running_tracks = num_tracks;
     song_looping = looping;
 
+    ticks_per_beat = MIDI_GetFileTimeDivision(file);
+
+    // Default is 120 bpm.
+    // TODO: this is wrong
+
+    us_per_beat = 500 * 1000;
+
     for (i=0; i<num_tracks; ++i)
     {
         StartTrack(file, i);
@@ -1377,7 +1411,7 @@ static void *I_OPL_RegisterSong(void *data, int len)
     {
         M_WriteFile(filename, data, len);
     }
-    else 
+    else
     {
 	// Assume a MUS file and try to convert
 
@@ -1394,8 +1428,7 @@ static void *I_OPL_RegisterSong(void *data, int len)
     // remove file now
 
     remove(filename);
-
-    Z_Free(filename);
+    free(filename);
 
     return result;
 }
@@ -1481,6 +1514,7 @@ music_module_t music_opl_module =
     I_OPL_PlaySong,
     I_OPL_StopSong,
     I_OPL_MusicIsPlaying,
+    NULL,  // Poll
 };
 
 //----------------------------------------------------------------------
@@ -1520,19 +1554,20 @@ static int ChannelInUse(opl_channel_data_t *channel)
     return 0;
 }
 
-void I_OPL_DevMessages(char *result)
+void I_OPL_DevMessages(char *result, size_t result_len)
 {
+    char tmp[80];
     int instr_num;
     int lines;
     int i;
 
     if (num_tracks == 0)
     {
-        sprintf(result, "No OPL track!");
+        M_snprintf(result, result_len, "No OPL track!");
         return;
     }
 
-    sprintf(result, "Tracks:\n");
+    M_snprintf(result, result_len, "Tracks:\n");
     lines = 1;
 
     for (i = 0; i < NumActiveChannels(); ++i)
@@ -1544,16 +1579,19 @@ void I_OPL_DevMessages(char *result)
 
         instr_num = tracks[0].channels[i].instrument - main_instrs;
 
-        sprintf(result + strlen(result),
-                "chan %i: %c i#%i (%s)\n",
-                i,
-                ChannelInUse(&tracks[0].channels[i]) ? '\'' : ' ',
-                instr_num + 1,
-                main_instr_names[instr_num]);
+        M_snprintf(tmp, sizeof(tmp),
+                   "chan %i: %c i#%i (%s)\n",
+                   i,
+                   ChannelInUse(&tracks[0].channels[i]) ? '\'' : ' ',
+                   instr_num + 1,
+                   main_instr_names[instr_num]);
+        M_StringConcat(result, tmp, result_len);
+
         ++lines;
     }
 
-    sprintf(result + strlen(result), "\nLast percussion:\n");
+    M_snprintf(tmp, sizeof(tmp), "\nLast percussion:\n");
+    M_StringConcat(result, tmp, result_len);
     lines += 2;
 
     i = (last_perc_count + PERCUSSION_LOG_LEN - 1) % PERCUSSION_LOG_LEN;
@@ -1564,11 +1602,12 @@ void I_OPL_DevMessages(char *result)
             break;
         }
 
-        sprintf(result + strlen(result),
-                "%cp#%i (%s)\n",
-                i == 0 ? '\'' : ' ',
-                last_perc[i],
-                percussion_names[last_perc[i] - 35]);
+        M_snprintf(tmp, sizeof(tmp),
+                   "%cp#%i (%s)\n",
+                   i == 0 ? '\'' : ' ',
+                   last_perc[i],
+                   percussion_names[last_perc[i] - 35]);
+        M_StringConcat(result, tmp, result_len);
         ++lines;
 
         i = (i + PERCUSSION_LOG_LEN - 1) % PERCUSSION_LOG_LEN;
