@@ -1,7 +1,5 @@
-// Emacs style mode select   -*- C++ -*- 
-//-----------------------------------------------------------------------------
 //
-// Copyright(C) 2005 Simon Howard
+// Copyright(C) 2005-2014 Simon Howard
 //
 // This program is free software; you can redistribute it and/or
 // modify it under the terms of the GNU General Public License
@@ -13,17 +11,14 @@
 // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 // GNU General Public License for more details.
 //
-// You should have received a copy of the GNU General Public License
-// along with this program; if not, write to the Free Software
-// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-// 02111-1307, USA.
-//
 // DESCRIPTION:
 //     Querying servers to find their current status.
 //
 
+#include <stdio.h>
 #include <stdarg.h>
 #include <stdlib.h>
+#include <string.h>
 
 #include "i_system.h"
 #include "i_timer.h"
@@ -44,6 +39,10 @@
 
 #define QUERY_TIMEOUT_SECS 2
 
+// Time to wait for secure demo signatures before declaring a timeout.
+
+#define SIGNATURE_TIMEOUT_SECS 5
+
 // Number of query attempts to make before giving up on a server.
 
 #define QUERY_MAX_ATTEMPTS 3
@@ -75,9 +74,8 @@ typedef struct
     boolean printed;
 } query_target_t;
 
-// Transmit a query packet
-
 static boolean registered_with_master = false;
+static boolean got_master_response = false;
 
 static net_context_t *query_context;
 static query_target_t *targets;
@@ -87,6 +85,8 @@ static boolean query_loop_running = false;
 static boolean printed_header = false;
 static int last_query_time = 0;
 
+static char *securedemo_start_message = NULL;
+
 // Resolve the master server address.
 
 net_addr_t *NET_Query_ResolveMaster(net_context_t *context)
@@ -150,9 +150,24 @@ void NET_Query_MasterResponse(net_packet_t *packet)
             printf("Failed to register with master server at %s\n",
                    MASTER_SERVER_ADDRESS);
         }
+
+        got_master_response = true;
     }
 }
 
+boolean NET_Query_CheckAddedToMaster(boolean *result)
+{
+    // Got response from master yet?
+
+    if (!got_master_response)
+    {
+        return false;
+    }
+
+    *result = registered_with_master;
+    return true;
+}
+
 // Send a query to the master server.
 
 static void NET_Query_SendMasterQuery(net_addr_t *addr)
@@ -526,10 +541,14 @@ static void NET_Query_QueryLoop(net_query_callback_t callback, void *user_data)
 
 void NET_Query_Init(void)
 {
-    query_context = NET_NewContext();
-    NET_AddModule(query_context, &net_sdl_module);
-    net_sdl_module.InitClient();
+    if (query_context == NULL)
+    {
+        query_context = NET_NewContext();
+        NET_AddModule(query_context, &net_sdl_module);
+        net_sdl_module.InitClient();
+    }
 
+    free(targets);
     targets = NULL;
     num_targets = 0;
 
@@ -640,23 +659,35 @@ static void formatted_printf(int wide, char *s, ...)
 
 static char *GameDescription(GameMode_t mode, GameMission_t mission)
 {
-    switch (mode)
+    switch (mission)
     {
-        case shareware:
-            return "shareware";
-        case registered:
-            return "registered";
-        case retail:
-            return "ultimate";
-        case commercial:
-            if (mission == doom2)
-                return "doom2";
-            else if (mission == pack_tnt)
-                return "tnt";
-            else if (mission == pack_plut)
-                return "plutonia";
+        case doom:
+            if (mode == shareware)
+                return "swdoom";
+            else if (mode == registered)
+                return "regdoom";
+            else if (mode == retail)
+                return "ultdoom";
+            else
+                return "doom";
+        case doom2:
+            return "doom2";
+        case pack_tnt:
+            return "tnt";
+        case pack_plut:
+            return "plutonia";
+        case pack_chex:
+            return "chex";
+        case pack_hacx:
+            return "hacx";
+        case heretic:
+            return "heretic";
+        case hexen:
+            return "hexen";
+        case strife:
+            return "strife";
         default:
-            return "unknown";
+            return "?";
     }
 }
 
@@ -797,3 +828,124 @@ net_addr_t *NET_FindLANServer(void)
     }
 }
 
+// Block until a packet of the given type is received from the given
+// address.
+
+static net_packet_t *BlockForPacket(net_addr_t *addr, unsigned int packet_type,
+                                    unsigned int timeout_ms)
+{
+    net_packet_t *packet;
+    net_addr_t *packet_src;
+    unsigned int read_packet_type;
+    unsigned int start_time;
+
+    start_time = I_GetTimeMS();
+
+    while (I_GetTimeMS() < start_time + timeout_ms)
+    {
+        if (!NET_RecvPacket(query_context, &packet_src, &packet))
+        {
+            I_Sleep(20);
+            continue;
+        }
+
+        if (packet_src == addr
+         && NET_ReadInt16(packet, &read_packet_type)
+         && packet_type == read_packet_type)
+        {
+            return packet;
+        }
+
+        NET_FreePacket(packet);
+    }
+
+    // Timeout - no response.
+
+    return NULL;
+}
+
+// Query master server for secure demo start seed value.
+
+boolean NET_StartSecureDemo(prng_seed_t seed)
+{
+    net_packet_t *request, *response;
+    net_addr_t *master_addr;
+    char *signature;
+    boolean result;
+
+    NET_Query_Init();
+    master_addr = NET_Query_ResolveMaster(query_context);
+
+    // Send request packet to master server.
+
+    request = NET_NewPacket(10);
+    NET_WriteInt16(request, NET_MASTER_PACKET_TYPE_SIGN_START);
+    NET_SendPacket(master_addr, request);
+    NET_FreePacket(request);
+
+    // Block for response and read contents.
+    // The signed start message will be saved for later.
+
+    response = BlockForPacket(master_addr,
+                              NET_MASTER_PACKET_TYPE_SIGN_START_RESPONSE,
+                              SIGNATURE_TIMEOUT_SECS * 1000);
+
+    result = false;
+
+    if (response != NULL)
+    {
+        if (NET_ReadPRNGSeed(response, seed))
+        {
+            signature = NET_ReadString(response);
+
+            if (signature != NULL)
+            {
+                securedemo_start_message = strdup(signature);
+                result = true;
+            }
+        }
+
+        NET_FreePacket(response);
+    }
+
+    return result;
+}
+
+// Query master server for secure demo end signature.
+
+char *NET_EndSecureDemo(sha1_digest_t demo_hash)
+{
+    net_packet_t *request, *response;
+    net_addr_t *master_addr;
+    char *signature;
+
+    master_addr = NET_Query_ResolveMaster(query_context);
+
+    // Construct end request and send to master server.
+
+    request = NET_NewPacket(10);
+    NET_WriteInt16(request, NET_MASTER_PACKET_TYPE_SIGN_END);
+    NET_WriteSHA1Sum(request, demo_hash);
+    NET_WriteString(request, securedemo_start_message);
+    NET_SendPacket(master_addr, request);
+    NET_FreePacket(request);
+
+    // Block for response. The response packet simply contains a string
+    // with the ASCII signature.
+
+    response = BlockForPacket(master_addr,
+                              NET_MASTER_PACKET_TYPE_SIGN_END_RESPONSE,
+                              SIGNATURE_TIMEOUT_SECS * 1000);
+
+    if (response == NULL)
+    {
+        return NULL;
+    }
+
+    signature = NET_ReadString(response);
+
+    NET_FreePacket(response);
+
+    return signature;
+}
+
