@@ -1,7 +1,5 @@
-// Emacs style mode select   -*- C++ -*- 
-//-----------------------------------------------------------------------------
 //
-// Copyright(C) 2005,2006 Simon Howard
+// Copyright(C) 2005-2014 Simon Howard
 //
 // This program is free software; you can redistribute it and/or
 // modify it under the terms of the GNU General Public License
@@ -13,16 +11,9 @@
 // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 // GNU General Public License for more details.
 //
-// You should have received a copy of the GNU General Public License
-// along with this program; if not, write to the Free Software
-// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-// 02111-1307, USA.
-//
-//-----------------------------------------------------------------------------
 //
 // Base interface that abstracts the text mode screen.
 //
-//-----------------------------------------------------------------------------
 
 #ifndef TXT_MAIN_H
 #define TXT_MAIN_H
@@ -32,9 +23,21 @@
 
 #include "txt_sdl.h"
 
+// textscreen key values:
+// Key values are difficult because we have to support multiple conflicting
+// address spaces.
+// First, Doom's key constants use 0-127 as ASCII and extra values from
+// 128-255 to represent special keys. Second, mouse buttons are represented
+// as buttons. Finally, we want to be able to support Unicode.
+//
+// So we define different ranges:
+// 0-255:    Doom key constants, including ASCII.
+// 256-511:  Mouse buttons and other reserved.
+// >=512:    Unicode values greater than 127 are offset up into this range.
+
 // Special keypress values that correspond to mouse button clicks
 
-#define TXT_MOUSE_BASE         0x10000
+#define TXT_MOUSE_BASE         256
 #define TXT_MOUSE_LEFT         (TXT_MOUSE_BASE + 0)
 #define TXT_MOUSE_RIGHT        (TXT_MOUSE_BASE + 1)
 #define TXT_MOUSE_MIDDLE       (TXT_MOUSE_BASE + 2)
@@ -42,6 +45,22 @@
 #define TXT_MOUSE_SCROLLDOWN   (TXT_MOUSE_BASE + 4)
 #define TXT_MAX_MOUSE_BUTTONS  16
 
+#define TXT_KEY_TO_MOUSE_BUTTON(x)                                        \
+        ( (x) >= TXT_MOUSE_BASE                                           \
+       && (x) < TXT_MOUSE_BASE + TXT_MAX_MOUSE_BUTTONS ?                  \
+          (x) - TXT_MOUSE_BASE : -1 )
+
+// Unicode offset. Unicode values from 128 onwards are offset up into
+// this range, so TXT_UNICODE_BASE = Unicode character #128, and so on.
+
+#define TXT_UNICODE_BASE       512
+
+// Convert a key value to a Unicode character:
+
+#define TXT_KEY_TO_UNICODE(x)                                             \
+        ( (x) < 128 ? (x) :                                               \
+          (x) >= TXT_UNICODE_BASE ? ((x) - TXT_UNICODE_BASE + 128) : 0 )
+
 // Screen size
 
 #define TXT_SCREEN_W 80
@@ -81,55 +100,55 @@ typedef enum
 
 // Initialize the screen
 // Returns 1 if successful, 0 if failed.
-
 int TXT_Init(void);
 
 // Shut down text mode emulation
-
 void TXT_Shutdown(void);
 
 // Get a pointer to the buffer containing the raw screen data.
-
 unsigned char *TXT_GetScreenData(void);
 
 // Update an area of the screen
-
 void TXT_UpdateScreenArea(int x, int y, int w, int h);
 
 // Update the whole screen
-
 void TXT_UpdateScreen(void);
 
 // Read a character from the keyboard
-
 int TXT_GetChar(void);
 
 // Read the current state of modifier keys that are held down.
-
 int TXT_GetModifierState(txt_modifier_t mod);
 
 // Provides a short description of a key code, placing into the 
 // provided buffer.
-
-void TXT_GetKeyDescription(int key, char *buf);
+void TXT_GetKeyDescription(int key, char *buf, size_t buf_len);
 
 // Retrieve the current position of the mouse
-
 void TXT_GetMousePosition(int *x, int *y);
 
 // Sleep until an event is received or the screen needs updating
 // Optional timeout in ms (timeout == 0 : sleep forever)
-
 void TXT_Sleep(int timeout);
 
 // Controls whether keys are returned from TXT_GetChar based on keyboard
 // mapping, or raw key code.
-
 void TXT_EnableKeyMapping(int enable);
 
 // Set the window title of the window containing the text mode screen
-
 void TXT_SetWindowTitle(char *title);
 
+// Safe string copy.
+void TXT_StringCopy(char *dest, const char *src, size_t dest_len);
+
+// Safe string concatenate.
+void TXT_StringConcat(char *dest, const char *src, size_t dest_len);
+
+// Safe version of vsnprintf().
+int TXT_vsnprintf(char *buf, size_t buf_len, const char *s, va_list args);
+
+// Safe version of snprintf().
+int TXT_snprintf(char *buf, size_t buf_len, const char *s, ...);
+
 #endif /* #ifndef TXT_MAIN_H */
 
